import argparse
import logging
import os
from abc import abstractmethod, ABC
from typing import Optional, Type

from cpk.cli.logger import cpklogger
from cpk.constants import CANONICAL_ARCH
from cpk.types import Machine, Arguments


class AbstractCLICommand(ABC):

    KEY = None

    @classmethod
    def name(cls) -> str:
        return cls.KEY

    @staticmethod
    def common_parser() -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument(
            "-C",
            "--workdir",
            default=os.getcwd(),
            help="Directory containing the CPK project"
        )
        parser.add_argument(
            "-H",
            "--machine",
            default=None,
            help="CPK machine or endpoint hostname where to perform the action"
        )
        parser.add_argument(
            "-a",
            "--arch",
            default=None,
            choices=set(CANONICAL_ARCH.values()),
            help="Target architecture for the image",
        )
        parser.add_argument(
            "-f",
            "--force",
            default=False,
            action="store_true",
            help="Whether to force the action",
        )
        parser.add_argument(
            "-v",
            "--verbose",
            default=False,
            action="store_true",
            help="Be verbose"
        )
        parser.add_argument(
            "--debug",
            default=False,
            action="store_true",
            help="Enable debug mode"
        )
        return parser

    @classmethod
    def get_parser(cls, args: Arguments) -> argparse.ArgumentParser:
        common_parser = cls.common_parser()
        command_parser = cls.parser(common_parser, args)
        command_parser.prog = f'cpk {cls.KEY}'
        return command_parser

    @staticmethod
    @abstractmethod
    def parser(parent: Optional[argparse.ArgumentParser] = None,
               args: Optional[Arguments] = None) -> argparse.ArgumentParser:
        pass

    @staticmethod
    @abstractmethod
    def execute(machine: Machine, parsed: argparse.Namespace) -> bool:
        pass


class CPKCLI:

    @staticmethod
    def parse_arguments(command: Type[AbstractCLICommand], args: Arguments) -> argparse.Namespace:
        parser = command.get_parser(args)
        parsed = parser.parse_args(args)
        # sanitize workdir
        parsed.workdir = os.path.abspath(parsed.workdir)
        # enable debug
        if parsed.debug:
            cpklogger.setLevel(logging.DEBUG)
        # ---
        return parsed


__all__ = [
    "AbstractCLICommand",
    "CPKCLI",
    "cpklogger"
]



import getpass
import dataclasses
import os
from typing import Optional


@dataclasses.dataclass
class ProjectVersion:
    tag: str
    head: Optional[str]
    closest: Optional[str]
    sha: Optional[str]


class GenericAdapter:

    def __init__(self, adapter_id: str):
        self._id = adapter_id
        self.enabled = False

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return "project"

    @property
    def registry(self) -> str:
        return os.environ.get("DOCKER_REGISTRY", "docker.io")

    @property
    def organization(self) -> str:
        return getpass.getuser()

    @property
    def description(self) -> Optional[str]:
        return None

    @property
    def version(self) -> ProjectVersion:
        return ProjectVersion(
            tag="latest",
            head=None,
            closest=None,
            sha=None
        )

    @property
    def maintainer(self) -> str:
        return getpass.getuser()

    @property
    def url(self) -> Optional[str]:
        return None

    def __str__(self) -> str:
        return self._id



import argparse
import glob
import json
import os
from pathlib import Path
from typing import Dict

import jsonschema

from cpk.cli import cpklogger
from cpk.schemas import get_machine_schema
from cpk.utils.misc import sanitize_hostname

from cpk.machine import Machine, FromEnvMachine, TCPMachine, SSHMachine, UnixSocketMachine

_supported_machines = {
    "ssh": SSHMachine,
    "tcp": TCPMachine,
    "socket": UnixSocketMachine
}


def load_machines(path: str) -> Dict[str, Machine]:
    machines = {}
    # iterate over the machines on disk
    for machine_cfg_fpath in glob.glob(os.path.join(path, '*/config.json')):
        machine_dir = Path(machine_cfg_fpath).parent
        machine_name = machine_dir.stem
        try:
            with open(machine_cfg_fpath, 'rt') as fin:
                try:
                    machine_cfg = json.load(fin)
                except json.decoder.JSONDecodeError as e:
                    raise ValueError(f"Machine descriptor file is not a valid JSON file. "
                                     f"Reason:\n\t{str(e)}")
                # make sure the key 'version' is present
                if "version" not in machine_cfg:
                    raise KeyError("Missing field 'version'.")
                # try reading the schema for this version
                try:
                    schema = get_machine_schema(machine_cfg["version"])
                except FileNotFoundError:
                    raise KeyError(f"Machine descriptor version '{machine_cfg['version']}' "
                                   f"not supported.")
                # validate config file
                try:
                    jsonschema.validate(machine_cfg, schema=schema)
                except jsonschema.exceptions.ValidationError as e:
                    raise ValueError(f"Machine descriptor has a bad format. "
                                     f"Reason:\n\t{str(e.message)}")
                # machine is valid, create object
                machine_cls = _supported_machines[machine_cfg["type"]]
                try:
                    machine = machine_cls(name=machine_name, **machine_cfg["configuration"])
                except TypeError as e:
                    raise ValueError(f"Machine descriptor has a bad format. "
                                     f"Reason:\n\t{str(e)}")
        except (KeyError, ValueError) as e:
            cpklogger.warning(f"An error occurred while loading the machine '{machine_name}', "
                              f"the error reads:\n{str(e)}")
            continue
        # we have loaded a valid machine
        machines[machine_name] = machine
    # ---
    return machines


def get_machine(parsed: argparse.Namespace, machines: Dict[str, Machine]) -> Machine:
    if parsed.machine is None:
        cpklogger.debug("Argument 'parsed.machine' not set. Creating machine from environment.")
        return FromEnvMachine()
    # match machine names against given string
    known_machine = machines.get(str(parsed.machine).strip(), None)
    if known_machine:
        cpklogger.debug(f"Machine '{parsed.machine}' is a known machine. "
                        f"Endpoint: {known_machine.base_url}")
        return known_machine
    # assume it is a hostname or IP address
    cpklogger.debug(f"Machine '{parsed.machine}' is not a known machine. "
                    f"Assuming a resolvable hostname or an IP address was passed.")
    return TCPMachine(parsed.machine, sanitize_hostname(parsed.machine))





import re
import termcolor as tc

from cpk.utils.misc import human_size

LAYER_SIZE_YELLOW = 20 * 1024 ** 2  # 20 MB
LAYER_SIZE_RED = 75 * 1024 ** 2  # 75 MB
SEPARATORS_LENGTH = 84
SEPARATORS_LENGTH_HALF = 25

EXTRA_INFO_SEPARATOR = '-' * SEPARATORS_LENGTH_HALF


class ImageAnalyzer:

    @staticmethod
    def process(buildlog, historylog, extra_info=None, nocolor=False):
        lines = buildlog

        # return if the log is empty
        if not lines:
            raise ValueError('The build log is empty')

        # return if the image history is empty
        if not historylog:
            raise ValueError('The image history is empty')

        if nocolor:
            tc.colored = lambda s, *_: s

        # define RegEx patterns
        step_pattern = re.compile("Step ([0-9]+)/([0-9]+) : (.*)")
        layer_pattern = re.compile(" ---> ([0-9a-z]{12})")
        cache_string = ' ---> Using cache'
        final_layer_pattern = re.compile("Successfully tagged (.*)")

        # sanitize log
        lines = list(map(lambda s: s.strip('\n'), lines))

        # check if the build process succeded
        if not final_layer_pattern.match(lines[-1]):
            return False
        image_names = []
        for line in reversed(lines):
            match = final_layer_pattern.match(line)
            if match:
                image_names.append(match.group(1))
            else:
                break

        # find "Step XY/TOT" lines
        steps_idx = [i for i in range(len(lines)) if step_pattern.match(lines[i])] + [len(lines)]

        # sanitize history log
        historylog = [
            (lid[7:19] if lid.startswith('sha256:') else lid, size) for (lid, size) in historylog
        ]

        # create map {layerid: size_bytes}
        layer_to_size_bytes = dict()
        for layerid, layersize in historylog:
            if 'missing' in layerid:
                continue
            layer_to_size_bytes[layerid] = int(layersize)

        # for each Step, find the layer ID
        first_layer = None
        cached_layers = 0
        for i, j in zip(steps_idx, steps_idx[1:]):
            indent_str = '|'
            layerid_str = 'Layer ID:'
            size_str = '   Size:'
            cur_step_lines = lines[i:j]
            open_layers = [
                layer_pattern.match(line) for line in cur_step_lines if layer_pattern.match(line)
            ]
            # check for cached layers
            step_cache = tc.colored('No', 'red')
            if first_layer is None or \
                    len(list(filter(lambda s: s == cache_string, cur_step_lines))) == 1:
                cached_layers += 1
                step_cache = tc.colored('Yes', 'green')
            # get Step info
            print('-' * SEPARATORS_LENGTH)
            stepline = lines[i]
            stepno = step_pattern.match(stepline).group(1)
            steptot = step_pattern.match(stepline).group(2)
            stepcmd = re.sub(' +', ' ', step_pattern.match(stepline).group(3))
            # get info about layer ID and size
            layerid = None
            layersize = 'ND'
            bg_color = 'white'
            fg_color = 'grey'
            if len(open_layers) > 0:
                layerid = open_layers[0].group(1)
                if stepcmd.startswith('FROM'):
                    first_layer = layerid
                    cached_layers += 1
            # ---
            if layerid in layer_to_size_bytes:
                layersize = human_size(layer_to_size_bytes[layerid])
                fg_color = 'white'
                bg_color = 'yellow' if layer_to_size_bytes[layerid] > LAYER_SIZE_YELLOW \
                    else 'green'
                bg_color = 'red' if layer_to_size_bytes[layerid] > LAYER_SIZE_RED else bg_color
                bg_color = 'blue' if stepcmd.startswith('FROM') else bg_color

            indent_str = tc.colored(indent_str, fg_color, 'on_' + bg_color)
            size_str = tc.colored(size_str, fg_color, 'on_' + bg_color)
            layerid_str = tc.colored(layerid_str, fg_color, 'on_' + bg_color)
            # print info about the current layer
            print(
                '%s %s\n%s   Step: %s/%s\n%s Cached: %s\n%sCommand: %s\n%s%s %s' % (
                    layerid_str, layerid,
                    indent_str, stepno, steptot,
                    indent_str, step_cache,
                    indent_str, stepcmd,
                    indent_str, size_str, layersize
                )
            )
            print()

        # get info about layers
        tot_layers = len(steps_idx) - 1
        cached_layers = min(tot_layers, cached_layers)

        # compute size of base and final image
        first_layer_idx = [
            i for i in range(len(historylog)) if historylog[i][0] == first_layer
        ][0]
        base_image_size = sum([int(line[1]) for line in historylog[first_layer_idx:]])
        final_image_size = sum([int(line[1]) for line in historylog])

        # print info about the whole image
        print()
        print(
            'Legend: %s %s\t%s %s\t%s < %s\t%s < %s\t%s > %s\t' % (
                tc.colored(' ' * 2, 'white', 'on_white'), 'EMPTY LAYER',
                tc.colored(' ' * 2, 'white', 'on_blue'), 'BASE LAYER',
                tc.colored(' ' * 2, 'white', 'on_green'),
                human_size(LAYER_SIZE_YELLOW, precision=1),
                tc.colored(' ' * 2, 'white', 'on_yellow'),
                human_size(LAYER_SIZE_RED, precision=1),
                tc.colored(' ' * 2, 'white', 'on_red'),
                human_size(LAYER_SIZE_RED, precision=1)
            )
        )
        print()
        print('=' * SEPARATORS_LENGTH)
        print('Final image name: %s' % ('\n' + ' ' * 18).join(image_names))
        print('Base image size: %s' % human_size(base_image_size))
        print('Final image size: %s' % human_size(final_image_size))
        print('Your image added %s to the base image.' % human_size(
            final_image_size - base_image_size))
        print(EXTRA_INFO_SEPARATOR)
        print('Layers total: {:d}'.format(tot_layers))
        print(' - Built: {:d}'.format(tot_layers - cached_layers))
        print(' - Cached: {:d}'.format(cached_layers))
        if extra_info is not None and len(extra_info) > 0:
            print(EXTRA_INFO_SEPARATOR)
            print(extra_info)
        print('=' * SEPARATORS_LENGTH)
        print()
        print(tc.colored('IMPORTANT', 'white', 'on_blue') +
              ': Always ask yourself, can I do better than that? ;)')
        print()
        # ---
        return image_names, base_image_size, final_image_size



import os
import json

import cpk

_SCHEMAS_DIR = os.path.join(os.path.dirname(cpk.__file__), 'schemas')


def _get_schema(schema_fpath: str) -> dict:
    if not os.path.isfile(schema_fpath):
        raise FileNotFoundError(schema_fpath)
    with open(schema_fpath, 'r') as fin:
        return json.load(fin)


def get_project_schema(schema: str) -> dict:
    schema_fpath = os.path.join(_SCHEMAS_DIR, "project.cpk", f"{schema}.json")
    return _get_schema(schema_fpath)


def get_template_schema(schema: str) -> dict:
    schema_fpath = os.path.join(_SCHEMAS_DIR, "template.cpk", f"{schema}.json")
    return _get_schema(schema_fpath)


def get_machine_schema(schema: str) -> dict:
    schema_fpath = os.path.join(_SCHEMAS_DIR, "machine", f"{schema}.json")
    return _get_schema(schema_fpath)


__all__ = [
    "get_project_schema",
    "get_template_schema",
    "get_machine_schema"
]



from .generic import GenericAdapter, ProjectVersion


class NullAdapter(GenericAdapter):

    def __init__(self, adapter_id: str):
        super(NullAdapter, self).__init__(adapter_id)
        self.enabled = True

    @property
    def name(self) -> None:
        return None

    @property
    def registry(self) -> None:
        return None

    @property
    def organization(self) -> None:
        return None

    @property
    def description(self) -> None:
        return None

    @property
    def version(self) -> ProjectVersion:
        return super(NullAdapter, self).version

    @property
    def maintainer(self) -> None:
        return None

    @property
    def url(self) -> None:
        return None



import logging
from logging import getLogger

from termcolor import colored

from cpk.constants import DEBUG

# create logger
cpklogger = getLogger('cpk')

colors = {
    'critical': 'red',
    'debug': 'magenta',
    'error': 'red',
    'info': None,
    'notice': 'magenta',
    'spam': 'green',
    'success': 'green',
    'verbose': 'blue',
    'warning': 'yellow'
}

# color parts of the left bar
levelname = colored("%(levelname)8s", "grey")
filename_lineno = colored("%(filename)15s:%(lineno)-4s", "blue")

# compile format
format = f"%(name)3s|{filename_lineno} - %(funcName)-15s : %(message)s" \
    if DEBUG else f"%(name)3s|{levelname} : %(message)s"
indent = " " * (43 if DEBUG else 13)


class CustomFilter(logging.Filter):
    def filter(self, record):
        lines = record.msg.split("\n")
        color = colors[record.levelname.lower()]
        lines = map(lambda l: colored(l, color) if color else l, lines)
        record.msg = f"\n{indent}: ".join(lines)
        return super(CustomFilter, self).filter(record)


# handle multi-line messages
cpklogger.addFilter(CustomFilter())

# create console handler
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
# create formatter and add it to the handlers
formatter = logging.Formatter(format)
ch.setFormatter(formatter)
# add the handlers to logger
cpklogger.addHandler(ch)


def update_logger(level: int):
    # set level
    cpklogger.setLevel(level)
    ch.setLevel(level)


# set INFO as default level
update_logger(logging.INFO)


__all__ = [
    "cpklogger",
    "update_logger"
]



import os
from pathlib import Path


DEBUG = False

CANONICAL_ARCH = {
    "arm": "arm32v7",
    "arm32v7": "arm32v7",
    "armv7l": "arm32v7",
    "armhf": "arm32v7",
    "x64": "amd64",
    "x86_64": "amd64",
    "amd64": "amd64",
    "Intel 64": "amd64",
    "arm64": "arm64v8",
    "arm64v8": "arm64v8",
    "armv8": "arm64v8",
    "aarch64": "arm64v8",
}

ARCH_TO_DOCKER_PLATFORM = {
    "amd64": "linux/amd64",
    "arm32v7": "linux/arm/v7",
    "arm64v8": "linux/arm64/v8",
}

BUILD_COMPATIBILITY_MAP = {
    "arm32v7": ["arm32v7"],
    "arm64v8": ["arm32v7", "arm64v8"],
    "amd64": ["amd64"]
}

CONTAINER_LABEL_DOMAIN = "cpk.label"

CPK_CONFIG_DIR = os.path.abspath(os.path.join(str(Path.home()), ".cpk"))

DOCKERHUB_API_URL = {
    "token": "https://auth.docker.io/token?scope=repository:{image}:pull&"
             "service=registry.docker.io",
    "digest": "https://registry-1.docker.io/v2/{image}/manifests/{tag}",
    "inspect": "https://registry-1.docker.io/v2/{image}/blobs/{digest}",
}



import abc
import copy
import dataclasses
import json
import logging
import os
from enum import Enum
from shutil import rmtree
from typing import List, Dict, Optional

import jsonschema
from docker import DockerClient

from cpk.utils.progress_bar import ProgressBar
from .constants import CANONICAL_ARCH
from .exceptions import \
    NotACPKProjectException, \
    InvalidCPKTemplateFile, InvalidCPKTemplate, \
    CPKTemplateSchemaNotSupported, CPKException
from .schemas import get_template_schema

Arguments = List[str]


@dataclasses.dataclass
class CPKProjectInfo:
    name: str
    organization: Optional[str]
    description: Optional[str]
    maintainer: Optional[str]
    template: 'CPKTemplateInfo'
    version: Optional[str]
    registry: Optional[str]
    tag: Optional[str]
    mappings: List['CPKFileMapping'] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class GitRepositoryVersion:
    head: Optional[str]
    closest: Optional[str]


@dataclasses.dataclass
class GitRepositoryOrigin:
    url: Optional[str]
    url_https: Optional[str]
    organization: Optional[str]


@dataclasses.dataclass
class GitRepositoryIndex:
    clean: bool
    num_added: Optional[int]
    num_modified: Optional[int]


@dataclasses.dataclass
class GitRepository:
    name: Optional[str]
    sha: Optional[str]
    branch: Optional[str]
    present: bool
    detached: bool
    version: GitRepositoryVersion
    origin: GitRepositoryOrigin
    index: GitRepositoryIndex

    @staticmethod
    def default() -> 'GitRepository':
        return GitRepository(
            name=None,
            sha=None,
            branch=None,
            present=False,
            detached=False,
            version=GitRepositoryVersion(
                head=None,
                closest=None
            ),
            origin=GitRepositoryOrigin(
                url=None,
                url_https=None,
                organization=None
            ),
            index=GitRepositoryIndex(
                clean=True,
                num_added=None,
                num_modified=None
            )
        )


@dataclasses.dataclass
class DockerImageRegistry:
    hostname: str = "docker.io"
    port: int = 5000

    def is_default(self) -> bool:
        defaults = DockerImageRegistry()
        # add - registry
        if self.hostname != defaults.hostname or self.port != defaults.port:
            return False
        return True

    def compile(self, allow_defaults: bool = False) -> Optional[str]:
        defaults = DockerImageRegistry()
        name = None if not allow_defaults else f"{defaults.hostname}"
        # add - registry
        if self.hostname != defaults.hostname:
            if self.port != defaults.port:
                name += f"{self.hostname}:{self.port}"
            else:
                name += f"{self.hostname}"
        # ---
        return name

    def __str__(self) -> str:
        return self.compile(allow_defaults=True)


@dataclasses.dataclass
class DockerImageName:
    """
    The official Docker image naming convention is:

        [REGISTRY[:PORT] /] USER / REPO [:TAG]

    """
    repository: str
    user: str = "library"
    registry: DockerImageRegistry = dataclasses.field(default_factory=DockerImageRegistry)
    tag: str = "latest"
    arch: Optional[str] = None

    def compile(self, allow_defaults: bool = False) -> str:
        name = ""
        defaults = DockerImageName("_")
        # add - registry
        registry = self.registry.compile(allow_defaults=allow_defaults)
        if registry:
            name += f"{registry}/"
        # add - user
        if self.user != defaults.user:
            name += f"{self.user}/"
        # add - repository
        name += self.repository
        # add - tag
        if self.tag != defaults.tag or self.arch:
            name += f":{self.tag}"
        # add - arch
        if self.arch:
            name += f"-{self.arch}"
        # ---
        return name

    @staticmethod
    def from_image_name(name: str) -> 'DockerImageName':
        input_parts = name.split('/')
        image = DockerImageName(
            repository="X"
        )
        # ---
        registry = None
        # ---
        if len(input_parts) == 3:
            registry, image.user, image_tag = input_parts
        elif len(input_parts) == 2:
            image.user, image_tag = input_parts
        elif len(input_parts) == 1:
            image_tag = input_parts[0]
        else:
            raise ValueError("Invalid Docker image name")
        image.repository, tag, *_ = image_tag.split(':') + ["latest"]
        for arch in set(CANONICAL_ARCH.values()):
            if tag.endswith(f"-{arch}"):
                tag = tag[:-(len(arch) + 1)]
                image.arch = arch
                break
        image.tag = tag
        if registry:
            image.registry.hostname, image.registry.port, *_ = registry.split(':') + [5000]
        # ---
        return image

    def __eq__(self, other):
        if isinstance(other, str):
            return other == self.compile() or other == self.compile(allow_defaults=True)
        elif isinstance(other, DockerImageName):
            return other.compile(allow_defaults=True) == self.compile(allow_defaults=True)
        return False

    def __str__(self) -> str:
        return f"""\
Registry:\t{str(self.registry)}
User:\t\t{self.user}
Repository:\t{self.repository}
Tag:\t\t{self.tag}
Arch:\t\t{self.arch}
        """


class CPKFileMappingTrigger(Enum):
    DEFAULT = "default"
    RUN_MOUNT = "run:mount"


@dataclasses.dataclass
class CPKFileMapping:
    source: str
    destination: str
    triggers: List[CPKFileMappingTrigger]
    required: bool
    mode: str = "rw"

    def __copy__(self):
        return self.__deepcopy__({})

    def __deepcopy__(self, memo):
        return CPKFileMapping(
            source=self.source,
            destination=self.destination,
            triggers=copy.deepcopy(self.triggers),
            required=self.required,
            mode=self.mode
        )

    @staticmethod
    def from_dict(data: dict) -> 'CPKFileMapping':
        return CPKFileMapping(
            source=data['source'],
            destination=data['destination'],
            triggers=list(map(CPKFileMappingTrigger, data.get('triggers', ["default"]))),
            required=data.get('required', False),
            mode=data.get('mode', "rw")
        )


@dataclasses.dataclass
class CPKTemplateInfo:
    name: str
    version: str
    mappings: List[CPKFileMapping]
    url: Optional[str]
    must_have: Dict[str, List[str]] = dataclasses.field(
        default_factory=lambda: {"files": [], "directories": []})

    @staticmethod
    def from_template_dict(data: dict) -> 'CPKTemplateInfo':
        # make sure the `schema` field is there
        if 'schema' not in data:
            raise InvalidCPKTemplate("Missing field: `schema`")
        # make sure we support that schema
        try:
            schema = get_template_schema(data["schema"])
        except FileNotFoundError:
            raise CPKTemplateSchemaNotSupported(data["schema"])
        # validate data against its declared schema
        try:
            jsonschema.validate(data, schema=schema)
        except jsonschema.ValidationError as e:
            raise InvalidCPKTemplate(str(e))
        # data is valid
        info = CPKTemplateInfo(
            name=data["name"],
            version=data["version"],
            mappings=list(map(lambda m: CPKFileMapping.from_dict(m), data.get('mappings', []))),
            url=data.get("url", None)
        )
        if "must_have" in data:
            info.must_have.update(data["must_have"])
        # ---
        return info

    @staticmethod
    def from_project_path(path: str) -> 'CPKTemplateInfo':
        metafile = os.path.join(path, "template.cpk")
        # check if the file 'template.cpk' is missing
        if not os.path.exists(metafile):
            raise NotACPKProjectException(path)
        # load 'template.cpk'
        with open(metafile, "rt") as fin:
            try:
                data = json.load(fin)
            except json.JSONDecodeError as e:
                raise InvalidCPKTemplateFile(path, f"File `{metafile}` must contain a valid JSON. "
                                                   f"DecoderError: {str(e)}")
        # make sure the `schema` field is there
        if 'schema' not in data:
            raise InvalidCPKTemplateFile(path, "Missing field: `schema`")
        # ---
        return CPKTemplateInfo.from_template_dict(data)


class Machine(abc.ABC):
    type: str

    def __init__(self, name: str, base_url: Optional[str] = None,
                 configuration: Optional[Dict] = None):
        self._name = name
        self._base_url = base_url
        self._configuration = configuration or {}
        # cache
        self._arch = None

    @property
    def name(self) -> str:
        return self._name

    @property
    @abc.abstractmethod
    def is_local(self) -> bool:
        pass

    @property
    def config_path(self) -> str:
        from cpk import cpkconfig
        return os.path.join(cpkconfig.path, "machines", self.name)

    @property
    def base_url(self) -> Optional[str]:
        return self._base_url

    @abc.abstractmethod
    def get_client(self) -> DockerClient:
        raise NotImplementedError("Method 'get_client' needs to be implemented by child class")

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def save(self, logger: Optional[logging.Logger] = None):
        from . import cpkconfig
        path = os.path.join(cpkconfig.path, "machines", self.name)
        # make directory
        if logger:
            logger.debug(f"Creating directory '{path}'")
        os.makedirs(path, mode=0o700, exist_ok=True)
        # dump config.json
        cfg_fpath = os.path.join(path, "config.json")
        if logger:
            logger.debug(f"Creating machine descriptor file '{cfg_fpath}'")
        with open(cfg_fpath, "wt") as fout:
            json.dump({
                "version": "1.0",
                "type": self.type,
                "description": "",
                "configuration": self._configuration
            }, fout, indent=4, sort_keys=True)
        os.chmod(cfg_fpath, 0o600)

    def remove(self, logger: Optional[logging.Logger] = None):
        from . import cpkconfig
        path = os.path.join(cpkconfig.path, "machines", self.name)
        if logger:
            logger.debug(f"Removing machine directory '{path}'")
        if os.path.exists(path):
            rmtree(path)

    def get_architecture(self) -> str:
        if self._arch is not None:
            return self._arch
        # get client
        client = self.get_client()
        try:
            # fetch architecture from client
            endpoint_arch = client.info()["Architecture"]
            if endpoint_arch not in CANONICAL_ARCH:
                raise CPKException(f"Unsupported architecture '{endpoint_arch}'.")
        finally:
            client.close()
        return CANONICAL_ARCH[endpoint_arch]

    def get_ncpus(self, ) -> Optional[int]:
        # noinspection PyBroadException
        try:
            return self.get_client().info()["NCPU"]
        except BaseException:
            return None

    def login_client(self):
        username = os.environ.get('DOCKER_USERNAME', None)
        password = os.environ.get('DOCKER_PASSWORD', None)
        if username is not None and password is not None:
            self.get_client().login(username=username, password=password)

    def pull_image(self, image, progress=True) -> bool:
        client = self.get_client()
        layers = set()
        pulled = set()
        pbar = ProgressBar() if progress else None
        for line in client.api.pull(image, stream=True, decode=True):
            if "id" not in line or "status" not in line:
                continue
            layer_id = line["id"]
            layers.add(layer_id)
            if line["status"] in ["Already exists", "Pull complete"]:
                pulled.add(layer_id)
            # update progress bar
            if progress:
                percentage = max(0.0, min(1.0, len(pulled) / max(1.0, len(layers)))) * 100.0
                pbar.update(percentage)
        if progress:
            pbar.done()
        return True

    def push_image(self, image, progress=True, **kwargs):
        client = self.get_client()
        layers = set()
        pushed = set()
        pbar = ProgressBar() if progress else None
        for line in client.api.push(*image.split(":"), stream=True, decode=True, **kwargs):
            if "id" not in line or "status" not in line:
                continue
            layer_id = line["id"]
            layers.add(layer_id)
            if line["status"] in ["Layer already exists", "Pushed"]:
                pushed.add(layer_id)
            # update progress bar
            if progress:
                percentage = max(0.0, min(1.0, len(pushed) / max(1.0, len(layers)))) * 100.0
                pbar.update(percentage)
        if progress:
            pbar.done()

    def __str__(self):
        return """
Machine:
    Type:\t{}
    Name:\t{}
    URL:\t{}
""".format(type(self).__name__, self.name, self.base_url)

    def __eq__(self, other):
        if not isinstance(other, Machine):
            return False
        return other.base_url == self.base_url


@dataclasses.dataclass
class CPKConfiguration:
    path: str
    machines: Dict[str, Machine]
